'''Parses the BeautifulSoup object from a Stack Exchange page.

Stack Exchange Q&A pages contain a three-level hierarchy of items:
questions, answers, and comments.
'''
from filum.helpers import (bs4_to_md, current_timestamp, iso_to_timestamp,
                           root_url)


def get_body(item):
    body = item.find('div', class_='s-prose js-post-body')
    return bs4_to_md(body)


def get_author(item):
    # The SE question author is last in the list of users on a question footer
    if item.find_all('div', class_='user-details')[-1].a:
        if item.find_all('div', class_='user-details')[-1].a.a:
            author = item.find_all('div', class_='user-details')[-1].a.a.string
        else:
            author = item.find_all('div', class_='user-details')[-1].a.string
    else:
        author = item.find_all('div', class_='user-details')[-1].span.string
    return author.strip()


def get_question_data(soup, url, site, question):

    title = [string for string in soup.find(id='question-header').stripped_strings][0]
    question_body = get_body(question)
    question_author = get_author(question)
    question_permalink = url + soup.find(id='question-header').h1.a.get('href')
    question_score = question.get('data-score')
    question_id = question.attrs['data-questionid']
    question_timestamp = soup.time.attrs['datetime']
    question_timestamp = iso_to_timestamp(question_timestamp)

    question_data = {
        'title': title,
        'body': question_body,
        'author': question_author,
        'id': question_id,
        'score': question_score,
        'permalink': question_permalink,
        'source': site,
        'posted_timestamp': question_timestamp,
        'saved_timestamp': current_timestamp()
    }

    return question_data


def parse_se(soup, url, site):
    'Wrapper function'
    soup = soup.find(id='content')
    root_is_answer = False
    if '#' in url:
        root_is_answer = True
        answer_id = url.split('#')[-1]
    url = root_url(url)
    question = soup.find(id='question')

    question_data = get_question_data(soup, url, site, question)

    children_data = {}

    def get_comments(comments):
        for comment in comments:
            comment_id = comment.get('data-comment-id')
            comment_score = comment.get('data-comment-score')
            comment_body = bs4_to_md(comment.find('span', class_='comment-copy')).replace('\n', '')
            comment_author = comment.select('.comment-user')[0].string
            comment_timestamp = comment.find('span', class_='relativetime-clean').attrs['title'].split('Z')[0]
            comment_timestamp = iso_to_timestamp(comment_timestamp)

            children_data.update({
                comment_id: {
                    'author': comment_author,
                    'text': comment_body,
                    'ancestor_id': question_data['id'],
                    'depth': 2,
                    'score': comment_score
                }
            })

    if not root_is_answer:
        # Don't bother extracting comments if the user wants to save a specific answer
        question_comments = question.find('div', class_='comments').find_all('li', class_='comment')
        get_comments(question_comments)

    if root_is_answer:
        answers = soup.find_all(id=f'answer-{answer_id}')
    else:
        answers = soup.find(id='answers').find_all('div', class_='answer')

    for answer in answers:
        answer_id = answer.get('data-answerid')
        answer_score = answer.get('data-score')
        answer_body = get_body(answer)
        answer_author = get_author(answer)
        answer_timestamp = answer.time.attrs['datetime']
        answer_timestamp = iso_to_timestamp(answer_timestamp)
        answer_permalink = url + '/a/' + answer_id

        children_data.update({
            answer_id: {
                'author': answer_author,
                'text': answer_body,
                'ancestor_id': question_data['id'],
                'depth': 1,
                'score': answer_score,
                'timestamp': answer_timestamp,
                'permalink': answer_permalink
            }
        })

        answer_comments = answer.find_all('li', class_='comment')
        if len(answer_comments) == 0:
            continue

        get_comments(answer_comments)

    thread = {
        'parent_data': question_data,
        'comment_data': children_data
    }

    return thread



import sys
import traceback

from rich.console import Console

from filum.download import Download
from filum.database import Database, ItemAlreadyExistsError
from filum.view import RichView

console = Console()


class Controller(object):
    def __init__(self):
        self.database = Database()
        self.view = RichView()

    def download_thread(self, url):
        return Download(url).run()

    def add_thread(self, thread):
        try:
            self.database.insert_row(thread['parent_data'], 'ancestors')
            for comment in thread['comment_data']:
                self.database.insert_row(thread['comment_data'][comment], 'descendants')
        except ItemAlreadyExistsError:
            # TODO: Allow updating of existing thread
            print('This item already exists in your database.')
            raise ItemAlreadyExistsError
            sys.exit(0)
        except Exception:
            traceback.print_exc()

    def update_thread(self, thread: dict):
        ancestor_id = self.database.update_ancestor(thread['parent_data'])
        self.database.delete_descendants(ancestor_id)
        for comment in thread['comment_data']:
            self.database.insert_row(thread['comment_data'][comment], 'descendants')

    def get_permalink(self, id: int) -> str:
        return self.database.select_permalink(id)

    def check_thread_exists(self, id):
        self.database.select_one_ancestor(id)

    def show_all_ancestors(self):
        results = self.database.select_all_ancestors()
        table = self.view.create_table(results)
        self.view.filum_print(table)

    def display_thread(self, id, pager, pager_colours, cond='', **kwargs):
        ancestor_query = self.database.select_one_ancestor(id, cond=cond, **kwargs)
        top_level = self.view.create_thread_header(ancestor_query)

        descendants_query = self.database.select_all_descendants(id, cond=cond, **kwargs)
        indented = self.view.create_thread_body(descendants_query)

        self.view.display_thread(top_level, indented, pager=pager, pager_colours=pager_colours)

    def delete(self, ancestor):
        self.database.delete_descendants(ancestor)
        self.database.delete_ancestor(ancestor)

    def get_ancestors_length(self):
        return self.database.get_ancestors_length()

    def modify_tags(self, id: int, add=True, **kwargs):
        '''Add or delete tags of a top-level item in the "ancestor" table
        :param int id: the ID of the item (in consecutive ascending order)
        :param bool add: default is to add tags, otherwise delete tags
        :key list tags: user-supplied tags to be added

        '''
        current_tags = self.database.get_tags(id)
        if current_tags is not None:
            current_tags = current_tags.split(', ')
        else:
            current_tags = []
        entered_tags = [tag.lower() for tag in kwargs['tags']]
        if add:
            # Ignore user-supplied tags that already exist
            new_tags = ', '.join(set(current_tags).union(entered_tags))
        else:
            new_tags = ', '.join([tag for tag in current_tags if tag not in entered_tags])
            if new_tags == '':
                new_tags = None
        self.database.update_tags(id, new_tags)

    def search(self, column, searchstr):
        results = self.database.search(column, searchstr)
        table = self.view.create_table(results)
        self.view.filum_print(table)



import argparse
import configparser
import pathlib
import platform
import subprocess
import sys
import textwrap
import warnings
from cmd import Cmd

from rich.console import Console

from filum.controller import Controller
from filum.database import ItemAlreadyExistsError
from filum.validation import InvalidInputError, is_valid_id, is_valid_url

console = Console()

parser = argparse.ArgumentParser(
                            description='Archive discussion threads',
                            prog='filum',
                            formatter_class=argparse.RawDescriptionHelpFormatter,
                            epilog=textwrap.dedent('''\
                            Example usage:

                            Add a URL
                            $ filum add <url>

                            View a table of all saved threads
                            $ filum all

                            Display a thread
                            $ filum show <thread label>
                            🖝 where <thread label> is the number in the left-most column of the table

                            Add tags to a saved thread
                            $ filum tags <tag 1> <tag 2> ... <tag n>
                            🖝 add the '--delete' flag to delete these tags instead
                            ''')
                                )

subparsers = parser.add_subparsers(dest='subparser')

parser_add = subparsers.add_parser('add', help='add a URL')
parser_add.add_argument('url', nargs='+', type=str, help='add a URL')
parser_add.set_defaults(parser_add=True)

parser_update = subparsers.add_parser('update', help='update a saved thread')
parser_update.add_argument('id', nargs=1, type=int)

parser_all = subparsers.add_parser('all', help='show all saved top-level items')
parser_all.set_defaults(parser_all=False)

parser_show = subparsers.add_parser('show', help='display a saved thread')
parser_show.add_argument('id', nargs=1, type=int)
parser_show.add_argument('--tags', nargs='+', help='display a thread selected from the table filtered by tags')
parser_show.add_argument('--source', nargs='+', help='display a thread selected from the table filtered by source')

parser_delete = subparsers.add_parser('delete', help='delete a saved thread')
parser_delete.add_argument('id', nargs='+', type=int)

parser_tags = subparsers.add_parser('tags', help='add tags. Include --delete to remove tags instead')
parser_tags.add_argument('id', nargs=1, type=int)
parser_tags.add_argument('tags', nargs='+', help='include one or more tags separated by a space')
parser_tags.add_argument('--delete', action='store_true')

parser_search = subparsers.add_parser('search', help='search for a thread')
parser_search.add_argument('--tags', nargs=1, help='filter table based on a tag')
parser_search.add_argument('--source', nargs=1, help='filter table by source')

parser_config = subparsers.add_parser('config', help='open config file')
parser_config.set_defaults(parser_config=False)

parser.add_argument('-i', action='store_true', help='interactive mode')
args = parser.parse_args()


def main():
    warnings.filterwarnings(
            'ignore',
            category=UserWarning,
            module='bs4',
            message='.*looks more like a filename than markup.*'
            )

    class FilumShell(Cmd):
        intro = 'filum interactive mode'
        prompt = 'filum > '

        def onecmd(self, line):
            try:
                return super().onecmd(line)
            except Exception as err:
                print(err)
                return False

        def emptyline(self):
            # Do nothing if an empty line is entered at the prompt
            pass

        def do_add(self, arg):
            '''Add a URL to the filum database: $ add <url>'''
            if arg == '':
                print('Please supply a URL.')
                return False
            add(arg)

        def do_update(self, arg):
            try:
                update(int(arg))
            except ValueError:
                print('Please enter a valid integer.')

        def do_all(self, arg):
            '''Show all top-level items currently saved in the filum database: $ all'''
            show_all()

        def do_show(self, line):
            '''Display a thread given its top-level selector: $ thread 1.\n
            Top-level selectors are contained in the left-most column in the table shown by the "all" command.'''
            args = parser_show.parse_args(line.split())
            try:
                if args.tags:
                    show_thread(args.id[0], cond='WHERE tags LIKE ?', where_param=f'%{args.tags[0]}%')
                elif args.source:
                    show_thread(args.id[0], cond='WHERE source LIKE ?', where_param=f'%{args.source[0]}%')
                else:
                    show_thread(args.id[0])
            except ValueError:
                print('Please enter a valid integer.')

        def do_delete(self, arg):
            '''Delete a thread given its top-level selector: $ thread 1.\n
            Top-level selectors are contained in the left-most column in the table shown by the "all" command.'''
            try:
                delete(int(arg))
            except ValueError:
                print('Please enter a valid integer.')

        def do_tags(self, line):
            try:
                args = parser_tags.parse_args(line.split())
                if args.delete:
                    modify_tags(args.id[0], add=False, tags=args.tags)
                else:
                    modify_tags(args.id[0], add=True, tags=args.tags)
            except SystemExit:
                return

        def do_search(self, line):
            try:
                args = parser_search.parse_args(line.split())
                if args.tags:
                    search('tags', args.tags[0])
                elif args.source:
                    search('source', args.source[0])
            except SystemExit:
                return

        def do_config(self, arg):
            '''Open the config file in an editor. Change settings by modifying the parameter values: $ config'''
            try:
                open_config()
            except Exception as err:
                print(err)

        def do_quit(self, arg):
            '''Quit the interactive session using 'quit' or CTRL-D'''
            sys.exit(0)

        def do_EOF(self, arg):
            '''Quit the interactive session using 'quit' or CTRL-D'''
            sys.exit(0)

    valid_id_message = 'Please enter a valid thread label (+ve int). Run `filum all` to see a list of thread labels.'

    config = configparser.ConfigParser()
    config_filepath = pathlib.Path(__file__).parent.resolve() / 'config.ini'
    config.read(config_filepath)

    c = Controller()

    def add(url) -> None:
        print(url)
        try:
            is_valid_url(url)
            with console.status(f'Downloading thread from {url}'):
                thread = c.download_thread(url)
                c.add_thread(thread)
            print('Thread downloaded.')
        except InvalidInputError as err:
            print(err)
        except ItemAlreadyExistsError:
            if confirm('Do you want to update this thread now? [y/n] '):
                print('Updating thread ...')
                c.update_thread(thread)

    def update(id: int) -> None:
        if confirm('Do you want to update this thread now? [y/n] '):
            with console.status('Updating thread...'):
                url = c.get_permalink(id)
                is_valid_url(url)
                thread = c.download_thread(url)
                c.update_thread(thread)
            print(f'Thread updated. ({url})')
            show_all()

    def show_thread(id: int, cond='', **kwargs) -> None:
        try:
            is_valid_id(id)
            c.display_thread(
                id,
                cond=cond,
                pager=config.getboolean('output', 'pager'),
                pager_colours=config.getboolean('output', 'pager_colours'),
                **kwargs
                )
        except InvalidInputError as err:
            print(err)
        except IndexError:
            print(valid_id_message)

    def show_all() -> None:
        c.show_all_ancestors()

    def delete(id: int) -> None:
        try:
            if confirm('Are you sure you want to delete this thread? [y/n] '):
                is_valid_id(id)
                ancestors_length = c.get_ancestors_length()
                success = True if id <= ancestors_length else False
                if success:
                    c.delete(id)
                    print(f'Thread no. {id} deleted.')
                    show_all()
                else:
                    print(f'Thread no. {id} does not exist.\n{valid_id_message}')
            else:
                print('Delete action cancelled.')
        except InvalidInputError as err:
            print(err)
        except IndexError:
            print(valid_id_message)

    def confirm(prompt) -> bool:  # type: ignore
        yes_no = ''

        while yes_no not in ('y', 'n'):
            yes_no = input(prompt)
            if yes_no == 'y':
                return True
            elif yes_no == 'n':
                return False
            else:
                print('Enter "y" for yes or "n" for no.')
                continue

    def modify_tags(id, add: bool, **kwargs):
        c.modify_tags(id, add, **kwargs)
        show_all()

    def search(column, searchstr):
        c.search(column, searchstr)

    def open_config():
        # filepath = pathlib.Path(__file__).parent.resolve() / 'config.ini'
        if platform.system() == 'Darwin':       # macOS
            subprocess.run(('open', config_filepath))
        elif platform.system() == 'Windows':    # Windows
            subprocess.run('notepad', config_filepath)
        else:                                   # Linux variants
            subprocess.run(('nano', config_filepath))

    description = (
        'filum - archive discussion threads from the command line.\n\n'
        'Usage:\n'
        'filum all\nfilum add <url>\nfilum thread <id>\nfilum delete <id>\n\n'
        'filum is a tool to save discussion threads from Reddit, Hacker News, and Stack Exchange on your PC. '
        'Like a bookmarking tool, but the text itself is saved locally. Worry no more about deleted threads.\n\n'
        'Run "filum -h" for a full list of options.'
    )

    if args.i:
        FilumShell().cmdloop()

    if args.subparser == 'config':
        print('Opening config file...')
        open_config()

    if args.subparser == 'add':
        add(args.url[0])

    elif args.subparser == 'update':
        update(args.id[0])

    elif args.subparser == 'all':
        show_all()

    elif args.subparser == 'show':
        if args.tags:
            show_thread(args.id[0], cond='WHERE tags LIKE ?', where_param=f'%{args.tags[0]}%')
        elif args.source:
            show_thread(args.id[0], cond='WHERE source LIKE ?', where_param=f'%{args.source[0]}%')
        else:
            show_thread(args.id[0])

    elif args.subparser == 'delete':
        delete(args.id[0])

    elif args.subparser == 'tags':
        if args.delete:
            modify_tags(args.id[0], add=False, tags=args.tags)
        else:
            modify_tags(args.id[0], add=True, tags=args.tags)

    elif args.subparser == 'search':
        if args.tags:
            search('tags', args.tags[0])
        elif args.source:
            search('source', args.source[0])

    else:
        print(description)


if __name__ == '__main__':
    main()



'''Parses the BeautifulSoup object created from the response from a
Hacker News item.

'''

from filum.helpers import current_timestamp, bs4_to_md, iso_to_timestamp


def parse_hn(obj):
    soup = obj.soup
    parent = soup.find('table', class_='fatitem')
    parent_id = parent.find('tr', class_='athing').attrs['id']
    parent_permalink = 'https://news.ycombinator.com/item?id=' + parent_id
    title = parent.find('a', class_='titlelink')
    parent_score = None
    children = soup.find_all('tr', class_='athing comtr')
    children_data = {}

    def get_comment_text(item):
        comment_field = item.find('span', class_='commtext')
        reply_span = comment_field.find('div', class_='reply')
        if reply_span:
            reply_span.decompose()
        # Call the prettify() method to ensure each tag is rendered on a separate line.
        # HN uses <p> tags to denote a new line.
        return bs4_to_md(comment_field.prettify())

    if title:
        title = title.contents[0]
        parent_score = parent.find('span', class_='score').contents[0]
        parent_score = int(parent_score.replace('points', ''))
        parent_body = None
        parent_author = parent.find('td', class_='subtext').find('a', class_='hnuser').contents[0]
    else:
        # If the root item doesn't have a title element, then
        # it's a comment
        title = parent.find('span', class_='onstory').a.contents[0]
        parent_body = get_comment_text(parent)
        parent_author = parent.find('a', class_='hnuser').contents[0]

    parent_timestamp = parent.find('span', class_='age').attrs['title']
    parent_timestamp = iso_to_timestamp(parent_timestamp)

    for child in children:
        depth = child.find('td', class_='ind').attrs['indent']
        author = child.find('a', class_='hnuser').contents[0]
        comment_id = child.attrs['id']
        comment_timestamp = child.find('span', class_='age').attrs['title']
        comment_timestamp = iso_to_timestamp(comment_timestamp)
        permalink = 'https://news.ycombinator.com/item?id=' + comment_id
        comment_body = get_comment_text(child)
        children_data.update({
            comment_id: {
                'author': author,
                'text': comment_body,
                'permalink': permalink,
                'ancestor_id': parent_id,
                'depth': depth,
                'timestamp': comment_timestamp
            }
        })

    parent_data = {
        'title': title,
        'body': parent_body,
        'author': parent_author,
        'id': parent_id,
        'score': parent_score,
        'source': obj.site,
        'permalink': parent_permalink,
        'posted_timestamp': parent_timestamp,
        'saved_timestamp': current_timestamp()
    }

    thread = {
        'parent_data': parent_data,
        'comment_data': children_data
    }

    return thread



from typing import Mapping, ValuesView, Any

from rich import box
from rich.console import Console, Group, group
from rich.markdown import Markdown
from rich.padding import Padding
from rich.table import Table
from rich.theme import Theme

from filum.helpers import timestamp_to_iso

console = Console(
    theme=Theme({'markdown.block_quote': 'yellow'}),
    style='on black')


class RichView:
    def __init__(self):
        self.console = console

    def stringify(self, row: ValuesView) -> tuple:
        '''Turns each item in the SQL query result into a string
        that can be passed to Table().add_row
        '''
        return tuple(str(i) for i in row)

    def create_table(self, row_list: list) -> Table:
        '''Construct a new table each time to prevent concatenating
        new tables together each time the "all" command is called in the
        interactive shell.
        '''
        table = Table(box=box.SIMPLE, expand=True)
        table.add_column('#', style='green')
        table.add_column('Title')
        table.add_column('Posted')
        table.add_column('Saved')
        table.add_column('Score')
        table.add_column('Source')
        table.add_column('Tags')

        # Convert each sqlite3.Row object to a dict
        rows = [dict(row) for row in row_list]

        for row in rows:
            row['posted_timestamp'] = timestamp_to_iso(row['posted_timestamp'])
            row['saved_timestamp'] = timestamp_to_iso(row['saved_timestamp'])
            table_row = (
                row['num'],
                row['title'],
                row['posted_timestamp'],
                row['saved_timestamp'],
                row['score'],
                row['source'],
                row['tags']
            )
            table.add_row(*self.stringify(table_row))
        return table

    def create_thread_header(self, item: Mapping) -> Group:
        timestamp = timestamp_to_iso(item['posted_timestamp'])
        to_print = (
            f'\n[bold bright_yellow]{item["author"]}[/bold bright_yellow] '
            f'[green]{item["score"]} pts[/green] [blue]{timestamp}[/blue] {item["permalink"]}\n\n'
            f'✎ {item["title"]}\n'
            )
        body: Any = ''
        if item['body']:
            body = Markdown(item["body"])
        top_level_group = Group(
            Padding(to_print, (0, 0, 0, 2)),
            Padding(body, (0, 0, 0, 2))
        )
        return top_level_group

    def create_thread_body(self, results: list) -> Group:
        @group()
        def make_panels(results: list):
            for result in results:
                text = Markdown(result['text'])
                timestamp = ''
                # Padding can only accept integers not floats
                indent = (result["depth"] + 2)*2
                if result['timestamp']:
                    timestamp = timestamp_to_iso(result['timestamp'])
                if result['score'] is not None:
                    score = f'{result["score"]} pts'
                else:
                    score = ''
                header = (
                    f'\n¬ [bold bright_cyan]{result["author"]}[/bold bright_cyan] '
                    f'[green]{score}[/green] [blue]{timestamp}[/blue]\n'
                    )

                yield Padding(header, (0, 2, 0, indent))
                yield Padding(text, (0, 2, 0, indent + 2))

        return make_panels(results)

    def display_thread(self, top_level, indented, pager=True, pager_colours=True) -> None:
        if not pager:
            self.filum_print(top_level)
            self.filum_print(indented)
        elif pager:
            with self.console.pager(styles=pager_colours):
                # Only works if terminal pager supports colour
                self.filum_print(top_level)
                self.filum_print(indented)

    def filum_print(self, item):
        self.console.print(item)



import traceback
import warnings

import requests
from bs4 import BeautifulSoup

from filum.parsers.parse_hn import parse_hn
from filum.parsers.parse_reddit import parse_reddit
from filum.parsers.parse_se import parse_se


class Download:
    def __init__(self, url):
        self.url = url

    def process_url(self):
        # Reddit's unofficial API - add a '.json' suffix to any url
        if 'reddit.com' in self.url:
            self.url += '.json'
            self.site = 'reddit'
        elif 'news.ycombinator.com' in self.url:
            self.site = 'hn'
        elif 'stackexchange.com' in self.url or \
             'stackoverflow.com' in self.url:
            self.site = 'se'

        return self

    def get_response(self):
        headers = {
            'user-agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:100.0) Gecko/20100101 Firefox/100.0',
            'dnt': '1',
            'accept-encoding': 'gzip, deflate, br',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
            'accept-language': 'en-US,en;q=0.5'
        }
        self.r = requests.get(self.url, headers=headers)
        return self

    def parse_html(self, raw: str):
        warnings.filterwarnings(
            'ignore',
            category=UserWarning,
            module='bs4',
            message='.*looks more like a filename than markup.*'
            )
        self.soup = BeautifulSoup(raw, features='html5lib')
        return self

    def prepare_response(self):
        if self.site == 'reddit':
            self.r = self.r.json()
        else:
            self.r = self.parse_html(self.r.text)
        return self

    def get_thread(self):
        if self.site == 'reddit':
            return parse_reddit(self)
        elif self.site == 'hn':
            return parse_hn(self)
        elif self.site == 'se':
            return parse_se(self.soup, self.url, self.site)

    def run(self):
        return self.process_url().get_response().prepare_response().get_thread()


hn_url_root = 'https://news.ycombinator.com/item?id=27336584'
hn_url_comment = 'https://news.ycombinator.com/item?id=31451536'
se_url_root = 'https://stats.stackexchange.com/questions/6/the-two-cultures-statistics-vs-machine-learning'
se_url_answer = 'https://stackoverflow.com/questions/15340582/python-extract-pattern-matches/15340666#15340666'
reddit_url_root = 'https://www.reddit.com/r/boardgames/comments/utkslk/tiny_epic_which_one_would_you_recommend/'


def main():
    Download(hn_url_root).process_url() \
                         .get_response() \
                         .prepare_response() \
                         .get_thread()


if __name__ == '__main__':
    try:
        main()
    except Exception:
        traceback.print_exc()



from filum.helpers import html_to_md, current_timestamp


def parse_reddit(obj):
    content = obj.r
    parent = content[0]['data']['children'][0]['data']
    comments = content[1]['data']['children']

    comment_data = {}

    def get_comments(comments_dict):
        for comment in comments_dict:
            id = comment['data']['name']  # Used as dict key for comments
            if comment['kind'] == 'more':
                # These are comments that are hidden under the fold. There is usually a link saying
                # '_load more comments_'
                # TODO: Rewrite reddit parser to hit new URL and retrieve unexpanded children
                # Depth will need to be checked programmatically.
                continue

            depth = comment['data']['depth']
            replies = comment['data']['replies']

            comment_body = comment['data']['body_html']
            comment_body = html_to_md(comment_body)
            comment_permalink = f'https://reddit.com{comment["data"]["permalink"]}'
            comment_data.update({
                id: {
                    'author': comment['data']['author'],
                    'text': comment_body,
                    'ancestor_id': parent['name'],
                    'depth': depth,
                    'score': comment['data']['score'],
                    'timestamp': comment['data']['created_utc'],
                    'permalink': comment_permalink,
                    }
                })
            if 'author_fullname' in comment['data'].keys():
                comment_data[id]['author_id'] = comment['data']['author_fullname']
            else:
                comment_data[id]['author_id'] = None

            if len(replies) == 0:
                continue
            else:
                get_comments(replies['data']['children'])

    get_comments(comments)
    body = html_to_md(parent['selftext_html']) if parent['selftext_html'] else None
    parent_permalink = f'https://www.reddit.com{parent["permalink"]}'  # The 'www' part is important
    parent_data = {
        'title': parent['title'],
        'body': body,
        'author': parent['author'],
        'id': parent['name'],
        'score': parent['score'],
        'permalink': parent_permalink,
        'source': obj.site,
        'posted_timestamp': parent['created_utc'],
        'saved_timestamp': current_timestamp()
    }

    thread = {
        'parent_data': parent_data,
        'comment_data': comment_data
    }

    return thread



import re
from collections import KeysView
from datetime import datetime

from markdownify import markdownify as md


def bs4_to_md(soup):
    return md(str(soup), heading_style='ATX')


def html_to_md(html):
    # TODO: Show full hyperlinks (hn truncates these)
    return md(md(html, heading_style='ATX'), heading_style='ATX')


def root_url(url):
    p = re.compile(r'https://.+\.com')
    return p.search(url).group(0)


def current_timestamp():
    return datetime.timestamp(datetime.now())


def iso_to_timestamp(time):
    return datetime.fromisoformat(time).timestamp()


def timestamp_to_iso(timestamp):
    return datetime.fromtimestamp(timestamp).isoformat(sep=' ', timespec='seconds')


def qmarks(sequence: KeysView) -> str:
    'Get a qmark SQL placeholder of arbitrary length.'
    return ', '.join(['?']*len(sequence))



import sqlite3
from sqlite3 import OperationalError, IntegrityError
import pathlib
from filum.helpers import qmarks, current_timestamp

db_name = pathlib.Path(__file__).parent.resolve() / 'filum'


class ItemAlreadyExistsError(Exception):
    pass


class Database(object):
    def __init__(self):
        self._conn = self.connect_to_db(db_name)
        # self._conn.set_trace_callback(print)
        self.sql = dict([
            ('ancestors_sequential', 'SELECT *, ROW_NUMBER() OVER (ORDER BY saved_timestamp DESC) as num FROM ancestors')  # noqa: E501
            ])

        with self._conn:
            self.create_table_ancestors()
            self.create_table_descendants()

    def connect_to_db(self, db=None):
        if db is None:
            my_db = ':memory:'
            print('New connection to in-memory SQLite db')
        else:
            my_db = f'{db}.db'
        conn = sqlite3.connect(my_db)
        # Return Row object from queries to allow accessing columns by name
        conn.row_factory = sqlite3.Row
        return conn

    def create_table_ancestors(self):
        with self._conn:
            sql = (
                'CREATE TABLE IF NOT EXISTS ancestors'
                '(row_id INTEGER PRIMARY KEY AUTOINCREMENT,'
                'id TEXT, title TEXT, body TEXT, posted_timestamp INTEGER, saved_timestamp INTEGER, '
                'score INTEGER, permalink TEXT UNIQUE, author TEXT, source TEXT,'
                'tags TEXT);'
            )

            try:
                self._conn.execute(sql)
            except OperationalError as err:
                print(err)

    def create_table_descendants(self):
        with self._conn:
            sql = '''CREATE TABLE IF NOT EXISTS descendants
                    (row_id INTEGER PRIMARY KEY AUTOINCREMENT,
                    ancestor_id INTEGER REFERENCES ancestors(row_id),
                    id TEXT,
                    parent_id TEXT,
                    text TEXT, permalink TEXT,
                    author TEXT, author_id TEXT,
                    score INTEGER, timestamp INTEGER,
                    depth INTEGER);'''
            try:
                self._conn.execute(sql)
            except OperationalError as err:
                print(err)

    def insert_row(self, thread: dict, table_name):
        with self._conn:
            columns = thread.keys()
            values = tuple(thread.values())
            to_insert = f'''({', '.join(columns)}) VALUES ({qmarks(columns)})'''
            sql = f'''INSERT INTO {table_name} ''' + to_insert
            try:
                self._conn.executemany(sql, (values,))
                self._conn.commit()
            except IntegrityError as err:
                # print(err)
                if 'UNIQUE' in str(err):
                    raise ItemAlreadyExistsError

    def update_ancestor(self, thread: dict) -> int:
        with self._conn:
            sql = 'UPDATE ancestors SET saved_timestamp = ?, score = ? WHERE permalink = ?'
            try:
                self._conn.execute(sql, (
                        current_timestamp(),
                        thread['score'],
                        thread['permalink']
                        )
                    )
                id: int = self._conn.execute(
                            ('SELECT ROW_NUMBER() OVER (ORDER BY saved_timestamp DESC) FROM ancestors WHERE permalink = ?'),  # noqa: E501
                            (thread['permalink'], )
                            ) \
                    .fetchone()[0]
            except OperationalError as err:
                print(err)
            return id

    def select_permalink(self, id: int) -> str:
        with self._conn:
            sql = f'WITH a AS ({self.sql["ancestors_sequential"]}) SELECT permalink FROM a WHERE num = ?'
            permalink = self._conn.execute(sql, (id, )).fetchone()[0]
            return permalink

    def select_all_ancestors(self):
        with self._conn:
            sql = self.sql['ancestors_sequential']
            results = self._conn.execute(sql).fetchall()

            return results

    def select_one_ancestor(self, id: int, cond='', **kwargs) -> sqlite3.Row:
        if 'where_param' in kwargs.keys():
            where_param = kwargs['where_param']
            execute_args = (where_param, id)
        else:
            execute_args = (id, )
        with self._conn:
            sql = f'WITH a AS ({self.sql["ancestors_sequential"]} {cond})SELECT * FROM a WHERE num = (?)'
            results = self._conn.execute(sql, execute_args).fetchone()
        return results

    def select_all_descendants(self, id: int, cond='', **kwargs) -> sqlite3.Row:
        if 'where_param' in kwargs.keys():
            where_param = kwargs['where_param']
            execute_args = (where_param, id)
        else:
            execute_args = (id, )
        with self._conn:
            sql = (
                'WITH joined AS ('
                'SELECT d.depth, d.row_id, d.score, d.timestamp, a.id, d.text, d.author, a.num AS key '
                'FROM descendants d '
                f'JOIN ({self.sql["ancestors_sequential"]} a {cond}) a '
                'ON d.ancestor_id = a.id) '
                'SELECT * FROM joined WHERE key = ?'
            )
            results = self._conn.execute(sql, execute_args).fetchall()
            return results

    def get_ancestors_length(self) -> int:
        with self._conn:
            sql = 'SELECT rowid FROM ancestors;'
            results = self._conn.execute(sql).fetchall()
            if results is not None:
                return len(results)
            else:
                return 0

    def delete_ancestor(self, id: int) -> None:
        with self._conn:
            sql_ancestors = '''
                                WITH a AS (
                                    SELECT id, ROW_NUMBER() OVER (ORDER BY saved_timestamp DESC) AS num FROM ancestors
                                )
                                DELETE FROM ancestors WHERE id IN (SELECT id FROM a WHERE num = ?)
                                '''
            self._conn.execute(sql_ancestors, (id,))

    def delete_descendants(self, id: int) -> None:
        with self._conn:
            sql_descendants = '''
                                WITH a AS (
                                    SELECT id, ROW_NUMBER() OVER (ORDER BY saved_timestamp DESC) AS num FROM ancestors
                                )
                                DELETE FROM descendants
                                WHERE ancestor_id IN (SELECT id FROM a WHERE num = ?);
                                '''
            self._conn.execute(sql_descendants, (id,))

    def get_tags(self, id: int) -> str:
        with self._conn:
            sql = (
                f'WITH a AS ({self.sql["ancestors_sequential"]}) '
                'SELECT tags FROM a WHERE num = ?'
                )

            tags = self._conn.execute(sql, (id, )).fetchone()[0]
        return tags

    def update_tags(self, id: int, tags: str):
        '''Update (either add or delete) tags based on user input'''
        with self._conn:
            sql = (
                f'WITH a AS ({self.sql["ancestors_sequential"]}) '
                'UPDATE ancestors SET tags = ? WHERE id IN (SELECT id FROM a WHERE num = ?)'
                )
            self._conn.execute(sql, (tags, id))

    def search(self, column: str, searchstr: str):
        param = f'%{searchstr}%'
        with self._conn:
            sql = (
                'SELECT ROW_NUMBER() OVER (ORDER BY saved_timestamp DESC) as num, '
                f'title, posted_timestamp, saved_timestamp, score, source, tags FROM ancestors WHERE {column} LIKE ?'
            )
            results = self._conn.execute(sql, (param, )).fetchall()
        return results


def main():

    db = Database()
    db.get_ancestors_length()


if __name__ == '__main__':
    main()



import re

url_pattern_reddit = re.compile(r'https:\/\/www.reddit.com\/r\/.+\/comments\/')
url_pattern_so = re.compile(r'https:\/\/stackoverflow.com\/((questions)|(q)|(a))')
url_pattern_se = re.compile(r'https:\/\/.+\.stackexchange.com\/((questions)|(q)|(a))')
url_pattern_hn = re.compile(r'https:\/\/news.ycombinator.com\/item')

# TODO: Add patterns for other SE sites such as Ask Ubuntu, Server Fault,
# Super User

url_patterns = [url_pattern_reddit, url_pattern_so,
                url_pattern_se, url_pattern_hn]

# Custom exceptions for input validation


class InvalidInputError(Exception):
    '''Exception for errors due to invalid user input'''


class InvalidUrl(InvalidInputError):
    '''Invalid URL'''
    def __init__(self):
        self.message = ('Please enter a URL prefixed with "https://".\n'
                        'Supported sites: Reddit, Hacker News, Stack Exchange')
        super().__init__(self.message)


class InvalidThreadId(InvalidInputError):
    '''Invalid thread ID'''
    def __init__(self):
        self.message = ('Please enter a valid thread ID (positive integer). '
                        'Run `filum all` to see a list of thread IDs.')
        super().__init__(self.message)


# Validation functions


def is_valid_url(arg: str) -> bool:
    url = arg.strip().lower()
    for pattern in url_patterns:
        if pattern.match(url):
            return True
    raise InvalidUrl


def is_valid_id(arg: int) -> bool:
    if type(arg) == int:
        if arg > 0:
            return True
    raise InvalidThreadId



import pathlib
import unittest
from test.expected_parser_outputs import expected_output_se

from bs4 import BeautifulSoup
from filum.parsers.parse_se import parse_se

html_fp = pathlib.Path(__file__).parent.resolve() / 'test_se.html'

with open(html_fp) as f:
    html_se = f.read()

soup_se = BeautifulSoup(html_se, 'html5lib')


class TestParsers(unittest.TestCase):
    def test_parse_se_returns_correct_dict(self):
        expected = expected_output_se
        output = parse_se(
            soup_se,
            'https://stackoverflow.com/questions/72349922/tableau-not-recognizing-snowflake-odbc-drivers-on-mac',
            'se'
            )
        self.assertCountEqual(output, expected)


if __name__ == '__main__':
    unittest.main()



import pathlib
import unittest

from bs4 import BeautifulSoup
from filum.download import Download

html_fp = pathlib.Path(__file__).parent.resolve() / 'test_se.html'

with open(html_fp) as f:
    html_se = f.read()


class TestDownload(unittest.TestCase):
    def test_parse_html_creates_soup_from_html(self):
        d = Download('test')
        d.parse_html(html_se)
        self.assertIsInstance(d.soup, BeautifulSoup)


if __name__ == '__main__':
    unittest.main()



import unittest

from filum.view import RichView

view = RichView()


class TestView(unittest.TestCase):
    def test_stringify(self):
        query_result = {
            'num': 1,
            'title': 'Some title',
            'posted_timestamp': 1622404036,
            'saved_timestamp': 1654005888.148201,
            'score': 0,
            'source': 'hn',
            'tags': None
            }
        result_values = query_result.values()

        stringified = ('1', 'Some title', '1622404036', '1654005888.148201', '0', 'hn', 'None')

        self.assertEqual(view.stringify(result_values), stringified)


if __name__ == '__main__':
    unittest.main()



import unittest
from filum import helpers
from bs4 import BeautifulSoup

html = '<h1>Test</h1>'


class TestHelpers(unittest.TestCase):

    def test_bs4_to_md(self):
        soup = BeautifulSoup(html, 'html.parser')
        md = helpers.bs4_to_md(soup)
        self.assertEqual(md, '# Test\n\n')

    def test_html_to_md(self):
        md = helpers.html_to_md(html)
        self.assertEqual(md, '# Test\n\n')

    def test_root_url(self):
        self.assertEqual(helpers.root_url('https://reddit.com/r/python'), 'https://reddit.com')

    def test_iso_to_timestamp(self):
        self.assertEqual(helpers.iso_to_timestamp('2022-06-02T11:16:50'), 1654165010)

    def test_timestamp_to_iso(self):
        self.assertEqual(helpers.timestamp_to_iso(1654165010), '2022-06-02 11:16:50')


if __name__ == '__main__':
    unittest.main()
